(*----------------------------------------------------------------------(C)-*)
(* Copyright (C) 2006-2016 Konstantin Korovin and The University of Manchester. 
   This file is part of iProver - a theorem prover for first-order logic.

   iProver is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 2 of the License, or 
   (at your option) any later version.
   iProver is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
   See the GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with iProver.  If not, see <http://www.gnu.org/licenses/>.         *)
(*----------------------------------------------------------------------[C]-*)


open Lib
open Options
open Statistics
open Logic_interface
open Instantiation_loop
open Resolution_loop


(*----- debug modifiable part-----*)

let dbg_flag = true

type dbg_gr = 
  | D_trace
  | D_mem
  | D_solve
  | D_shared_clauses
  | D_eq_types
  | D_reset 

let dbg_gr_to_str = function 
  | D_trace -> "true"	
  | D_mem -> "mem"
  | D_solve -> "solve"
  | D_shared_clauses -> "shared_clauses"
  | D_eq_types -> "eq_types"
  | D_reset -> "reset"

let dbg_groups = [
   D_trace;  
  D_reset;
  (* D_mem; *)
  (* D_solve; *)
(*  D_shared_clauses; *)
(*  D_eq_types; *)
  
  D_eq_types;
]
    
let module_name = "proof_search_loop"

(*----- debug fixed part --------*)

let () = Lib.dbg_flag_msg dbg_flag module_name

let dbg group str_lazy =
  Lib.dbg_out_pref dbg_flag dbg_groups group dbg_gr_to_str module_name str_lazy

let dbg_env group f =
  Lib.dbg_env_set dbg_flag dbg_groups group f

(*----- debug -----*)



module PropSolver = Prop_solver_exchange.PropSolver

(* change Sched_Time_Out to PS_loop_timeout *)

exception PS_loop_time_out of int

let check_sched_time ~start_time ~time_limit full_loop_counter =
  match time_limit with
  | Undef -> ()
  | Def time_limit ->
    if Float.O.(Unix.gettimeofday () -. start_time > time_limit)
    then raise (PS_loop_time_out full_loop_counter)
    else ()


let time_to_string time =
  match time with
  | Def(float) -> string_of_float float
  | Undef -> "Unbounded"


(*------------ proof search state --------------------*)
        

type ps_input_clauses = 
    {
     ps_clauses_with_eq_axioms : clause list;
     ps_clauses_no_eq_axioms   : clause list;
     ps_problem_properties     : Problem_properties.prob_props; (* currently used only to check if problem is equational *)
   }

let clauses_with_eq_axs_to_ps_input_clauses clause_list = 
  {
   ps_clauses_with_eq_axioms=clause_list;
   ps_clauses_no_eq_axioms=clause_list;
   ps_problem_properties= Problem_properties.get_prob_props clause_list;
 }




type ps_state = 
    {
     mutable ps_input_clauses : ps_input_clauses;
     mutable ps_loop_cnt      : int;

     mutable ps_res_state  : res_state param; 
     mutable ps_res_cnt    : int; 

     mutable ps_inst_state : inst_state param;
     mutable ps_inst_cnt    : int; 
     mutable ps_learning_bound : int;
     mutable ps_learning_counter : int;

     mutable ps_sup_state  : Superposition.state option;
     mutable ps_sup_cnt    : int; 

     mutable ps_opts       : Options.options;
     
   }

(*
let ps_create_state () = 
  {
   ps_res_state = Def(res_create_state ~res_prep_only:false);
   ps_inst_state = Def(inst_create_state ());
 }
*)

(* TODO add option *)

let res_create_state_shared ~opts = 
   let res_state = res_create_state ~opts ~res_prep_only:false in
   (if opts.share_sel_clauses
   then
     (let shared_clauses = BCSet.elements (Shared_clauses.get_all_shared_clauses ()) in 
     dbg D_shared_clauses (lazy ("res adding: "^(Clause.clause_list_to_string shared_clauses)));
     res_add_clause_list res_state shared_clauses;
     )
   );
   res_state

let inst_create_state_shared ~opts = 
  let inst_state = inst_create_state ~opts in
  (if opts.share_sel_clauses
  then     
    (let shared_clauses = BCSet.elements (Shared_clauses.get_all_shared_clauses ()) in 
    dbg D_shared_clauses (lazy ("inst adding: "^(Clause.clause_list_to_string shared_clauses)));
    inst_add_clause_list inst_state shared_clauses;
    )
  );
  inst_state

let ps_create_state ~input_clauses ~opts = 
  (* Recompute problem properties *)
  (* let input_clauses = {input_clauses with ps_problem_properties = Problem_properties.get_prob_props input_clauses.ps_clauses_no_eq_axioms} in *)
{
  ps_input_clauses = input_clauses;
  ps_loop_cnt = 0;

  ps_res_state = Def(res_create_state_shared ~opts);
  ps_res_cnt   = 0; 
  
  ps_inst_state = Def(inst_create_state_shared ~opts);
  ps_inst_cnt   = 0; 
  ps_learning_bound   = opts.inst_learning_start;
  ps_learning_counter = 0; 

  (* ps_inst_restart_cnt = 0; *)

  ps_opts = Options.copy_options opts;

  ps_sup_cnt = 0; 

  ps_sup_state = if opts.superposition_flag then (
    (* Problem properties *)
    let prob_props = input_clauses.ps_problem_properties in
    (* let demod_use_ground = opts.demod_use_ground in *)

    (* Selectively disable demod and subsumption (disabling subsumption deprecated due to unit index) *)
    let demod_flag = Problem_properties.has_eq prob_props in
    let subs_flag = true (* not @@ Problem_properties.is_unit_eq prob_props *) in
    let sim_flag = 
      match demod_flag, subs_flag with
      | true, true -> Superposition.AllSims
      | false, _ -> Superposition.NoDemod
      | true, false -> Superposition.NoSubs
    in

    (* Discover eq_types *)
    let eq_types = 
      let signature = Clause.clause_list_signature input_clauses.ps_clauses_no_eq_axioms in
      dbg D_eq_types @@ lazy (sprintf "sig_eq_types:\n%s"
        (list_to_string Symbol.to_string (SSet.elements signature.sig_eq_types) "\n")
      );
      dbg D_eq_types @@ lazy (sprintf "sig_pure_dis_eq_types:\n%s"
        (list_to_string Symbol.to_string (SSet.elements signature.sig_pure_dis_eq_types) "\n")
      );
      SSet.diff signature.sig_eq_types signature.sig_pure_dis_eq_types
    in
    dbg D_eq_types @@ lazy (sprintf "Eq types:\n%s"
      (list_to_string Symbol.to_string (SSet.elements eq_types) "\n")
    );
    dbg_env D_eq_types (fun () -> 
      (* if not (SSet.elements eq_types = [Symbol.symb_default_type]) then  *)
      if not @@ List.X.equal ~eq:(==) (SSet.elements eq_types) [Symbol.symb_default_type] then 
        dbg D_eq_types @@ lazy "There exist some non-equality types"
    );

    let sup_options = Superposition.make_options opts ~sim_flag ~eq_types ~prob_props in
    let sup_state = Superposition.create sup_options in
    Some sup_state
  ) else None;
}


let ps_get_inst_state ps_state = get_param_val ps_state.ps_inst_state
let ps_get_res_state ps_state  = get_param_val ps_state.ps_res_state

let ps_inst_clear ps_state = ps_state.ps_inst_state <- Undef
let ps_res_clear ps_state  = ps_state.ps_res_state <- Undef

let ps_clear_state ps_state = 
  ps_inst_clear ps_state;
  ps_res_clear ps_state


(* --------- resets -----------*)

(*
let ps_reset ~reset_fun ps_state = 

  dbg_env D_mem 
    (fun () -> 
      print_objsize "ps_reset_inst_state:ps_state" ps_state
    );
(*  if not dbg_flag then reset_fun (); *)
  reset_fun ps_state;

  dbg_env D_mem 
    (fun () -> 
      print_mem_time_fun reset_fun ();
      out_basic_mem () 
    )
  *)
    
let ps_reset_inst_state ~opts ps_state =    
(*
  let reset_fun () = 
    Prop_solver_exchange.clear_soft_assumptions ();
    ps_state.ps_inst_state <- Def(inst_create_state ~opts);
    inst_add_clause_list (ps_get_inst_state ps_state) ps_state.ps_input_clauses.ps_clauses_with_eq_axioms;
  in  
  ps_reset ~reset_fun ps_state
  *)
  dbg D_reset (lazy ("ps_reset_inst_state"));
  Prop_solver_exchange.clear_soft_assumptions ();

  ps_state.ps_inst_state <- Def(inst_create_state ~opts);

  inst_add_clause_list (ps_get_inst_state ps_state) ps_state.ps_input_clauses.ps_clauses_with_eq_axioms

  
let ps_reset_res_state ~opts ps_state = 
(*
  let reset_fun () = 
    Prop_solver_exchange.clear_soft_assumptions ();
    ps_state.ps_res_state <- Def(res_create_state ~opts ~res_prep_only:false);
    res_add_clause_list (ps_get_res_state ps_state) ps_state.ps_input_clauses.ps_clauses_with_eq_axioms; 
  in  
  ps_reset ~reset_fun ps_state
*)
  Prop_solver_exchange.clear_soft_assumptions ();
  ps_state.ps_res_state <- Def(res_create_state ~opts ~res_prep_only:false);
  res_add_clause_list (ps_get_res_state ps_state) ps_state.ps_input_clauses.ps_clauses_with_eq_axioms
  
    
let ps_reset_state ~opts ps_state = 
  ps_reset_inst_state ~opts ps_state;
  ps_reset_res_state ~opts ps_state (* KK: do not reset res state ?*)

(*
let ps_reset_state ~opts ps_state = 
  dbg_env D_mem 
    (fun () -> 
      print_objsize "ps_state" ps_state
    );
  Prop_solver_exchange.clear_soft_assumptions ();
  let reset_fun () = 
    ps_state.ps_inst_state <- Def(inst_create_state ~opts);
    ps_state.ps_res_state  <- Def(res_create_state ~opts ~res_prep_only:false);
  in  
  if not dbg_flag then reset_fun ();
  dbg_env D_mem 
    (fun () -> 
      print_mem_time_fun reset_fun ();
      out_basic_mem () 
    )
  *)
   
 


(* last printed options *)
let last_used_options = ref None

(*KK clean!! *)

(* set options and print them *)
let save_and_print_options ~opts =
  (* save current options *)
  last_used_options := Some opts;
  (* print options *)
  (* out_str ((s_pref_str ())^"Current options:\n"^(options_to_str opts)^"\n\n"); *)
  out_str (s_pref_str ());
  print_string "Current options:\n";
  options_to_stdout opts;
  print_string "\n\n";
  (* that's it *)
  ()

(* print options if they are changed *)
let print_options_if_new ~opts =
  (* if Option.equal ~eq:Poly.(=) !last_used_options opts then *)
  match !last_used_options with
  (* same options -- don't print *)
  | Some op when Poly.(op = opts) -> ()
  (* not the same options -- save and print *)
  | _ -> save_and_print_options ~opts


let prop_solver_per_cl_check ~opts = 
  if (opts.prop_solver_per_cl > 0 && 
      ((Statistics.get_val_stat Statistics.prop_num_of_clauses) >= 
      (opts.prop_solver_per_cl * (Statistics.get_val_stat_fun Statistics.prop_solver_calls))))
  then
    (if Prop_solver_exchange.solve ~reset:false () == PropSolver.Unsat
    then
      (* Raise separate exception, solver is not in an invalid state
	 and can be satisfiable without assumptions *)
      (* raise PropSolver.Unsatisfiable *)
      raise Unsatisfiable_gr
    )  
  else
    ()

(*-------- total number of clauses in inst+rest+sup -----------*)

let get_inst_num_of_clauses () = 
  try
    Statistics.get_val_stat_fun inst_num_of_clauses
  with _ -> 0 
    
let get_res_num_of_clauses () = 
  try 
    Statistics.get_val_stat_fun res_num_of_clauses
  with _ -> 0 

let get_sup_num_of_clauses () = 
  try
    Statistics.get_val_stat_fun sup_num_of_clauses
  with _ -> 0 

let get_total_num_of_clauses () = 
  (get_inst_num_of_clauses ()) + (get_res_num_of_clauses ()) + (get_sup_num_of_clauses ())

(* is_turn is based on the proportion of clauses in section: inst/res/sup among all clauses *)

let is_turn ~opts ~mult ~sec_num_of_cl = 
  let mult_sum = opts.comb_inst_mult + opts.comb_res_mult + opts.comb_sup_mult in   
  assert(mult_sum > 0);
  let total_cl = get_total_num_of_clauses () in 
  dbg D_trace @@ lazy (sprintf "is_turn: total_cl %d; mult: %d; mut_sum: %d; cond_cl: %d; sec_num_of_cl: %d" total_cl mult mult_sum ((mult * total_cl)/mult_sum) sec_num_of_cl);
  
  ((mult * total_cl)/mult_sum) >= sec_num_of_cl

let is_inst_turn ~opts =
  dbg D_trace @@ lazy ("is_inst_turn");
  is_turn ~opts ~mult:(opts.comb_inst_mult) ~sec_num_of_cl:(get_inst_num_of_clauses ())

let is_res_turn ~opts = 
  dbg D_trace @@ lazy ("is_res_turn");
  is_turn ~opts ~mult:(opts.comb_res_mult) ~sec_num_of_cl:(get_res_num_of_clauses ())

let is_sup_turn ~opts = 
  dbg D_trace @@ lazy ("is_sup_turn");
  is_turn ~opts ~mult:(opts.comb_sup_mult) ~sec_num_of_cl:(get_sup_num_of_clauses ())
  

let large_neg_const = -100000000 

let turn_deficit ~opts ~mult ~sec_num_of_cl = 
  let mult_sum = 
    (if opts.instantiation_flag then opts.comb_inst_mult else large_neg_const) + 
    (if opts.resolution_flag then  opts.comb_res_mult else large_neg_const) + 
    (if opts.superposition_flag then opts.comb_sup_mult else large_neg_const) in   
  assert(mult_sum > 0);
  let total_cl = get_total_num_of_clauses () in 
  let turn_deficit = ((mult * total_cl)/mult_sum) - sec_num_of_cl in

  dbg D_trace @@ lazy (sprintf "is_turn: total_cl %d; mult: %d; mut_sum: %d; cond_cl: %d; sec_num_of_cl: %d; turn_deficit" total_cl mult mult_sum sec_num_of_cl turn_deficit);
  turn_deficit

let inst_turn_deficit ~opts =
  dbg D_trace @@ lazy ("is_inst_turn");
  turn_deficit ~opts ~mult:(opts.comb_inst_mult) ~sec_num_of_cl:(get_inst_num_of_clauses ())

let res_turn_deficit  ~opts = 
  dbg D_trace @@ lazy ("is_res_turn");
  turn_deficit ~opts ~mult:(opts.comb_res_mult) ~sec_num_of_cl:(get_res_num_of_clauses ())
    
let sup_turn_deficit ~opts = 
  dbg D_trace @@ lazy ("is_sup_turn");
  turn_deficit ~opts ~mult:(opts.comb_sup_mult) ~sec_num_of_cl:(get_sup_num_of_clauses ())
  



  
(*
let is_inst_turn ~opts = 
  let mult_sum = opts.comb_inst_mult + opts.comb_res_mult + opts.comb_sup_mult in   
  assert(mult_sum > 0);
  let total_cl = get_total_num_of_clauses () in 
  let inst_num_of_clauses = get_inst_num_of_clauses () in 
  ((opts.comb_inst_mult * total_cl)/mult_sum) >= inst_num_of_clauses
*)


    
(*----------------------Full Loop--------------------------------------*)

let ps_full_loop ~opts ~time_limit input_clauses =
  let o = opts in  
  let start_time = (Unix.gettimeofday ()) in
  
  print_options_if_new ~opts;
  
  out_str "\n";
  print_string ((s_pref_str ())^"Proving...\n");
  flush stdout;

  let ps_state = ps_create_state ~input_clauses ~opts:o in
  dbg D_trace (lazy ("adding clauses to ps_state: "));
  dbg D_trace (lazy (Clause.clause_list_to_string input_clauses.ps_clauses_with_eq_axioms));
  (if o.instantiation_flag then inst_add_clause_list (ps_get_inst_state ps_state) input_clauses.ps_clauses_with_eq_axioms); 
  (if o.resolution_flag    then res_add_clause_list (ps_get_res_state ps_state) input_clauses.ps_clauses_with_eq_axioms);  
  (if o.superposition_flag then 
    Superposition.add_input_clauses (ps_state.ps_sup_state |> Option.get) ((* EqualityTransformation.clauselist_to_eq *) input_clauses.ps_clauses_no_eq_axioms);
  );  
  dbg D_trace (lazy ("done adding clauses "));

  let learning_bound = ref o.inst_learning_start in
  let learning_counter = ref 0 in
  let resolution_counter = ref 0 in
  let instantiation_counter = ref 0 in
  let superposition_counter = ref 0 in
  let full_loop_counter = ref 0 in

  let previous_total_num_of_clauses = ref 0 in 
  let num_of_clauses_unchanged = ref false in
  let inst_turn_deficit_ref = ref 0 in 
  let res_turn_deficit_ref = ref 0 in 
  let sup_turn_deficit_ref = ref 0 in 

  while true do
    (     

     num_of_clauses_unchanged := 
      (dbg D_trace @@ lazy (sprintf "prev_tot_cl: %d; curr_tot_cl %d" !previous_total_num_of_clauses (get_total_num_of_clauses ()));
(*       !previous_total_num_of_clauses = (get_total_num_of_clauses ()); *)
(* set false for now *)
       false
);
       
     previous_total_num_of_clauses := get_total_num_of_clauses ();

     inst_turn_deficit_ref := inst_turn_deficit ~opts;
     res_turn_deficit_ref  := res_turn_deficit ~opts; 
     sup_turn_deficit_ref := sup_turn_deficit ~opts;

     let max_turn_deficit = list_find_max_element Poly.compare 
         [if o.instantiation_flag then !inst_turn_deficit_ref else large_neg_const;
          if o.resolution_flag then !res_turn_deficit_ref else large_neg_const;
          if o.superposition_flag then !sup_turn_deficit_ref else large_neg_const] 
     in
     dbg D_trace @@ lazy (sprintf "max_turn_deficit: %d" max_turn_deficit);
(*          
     let num_of_clauses_is_unchanged () = (
       dbg D_trace @@ lazy (sprintf "prev_tot_cl: %d; curr_tot_cl %d" !previous_total_num_of_clauses (get_total_num_of_clauses ()));
       !previous_total_num_of_clauses = (get_total_num_of_clauses ())
      ) in
*)
     check_sched_time ~start_time ~time_limit !full_loop_counter;
     full_loop_counter:= !full_loop_counter +1;
     if (not o.resolution_flag) && (not o.instantiation_flag) && (not o.superposition_flag) then failwith "both --resolution_flag false and --instantiation_flag and --superposition_flag false";

     prop_solver_per_cl_check ~opts;

(*     if (o.comb_res_mult = 0) && (o.comb_inst_mult =0) then failwith "both comb_res_mult=0  and comb_inst_mult=0"; *)
(*
     dbg D_trace (lazy (" loop: "^(string_of_int !full_loop_counter)));

     dbg D_trace (lazy ("o.resolution_flag: "^(string_of_bool o.resolution_flag)
                        ^" o.comb_res_mult: "^(string_of_int o.comb_res_mult)
                        ^" !resolution_counter: "^(string_of_int !resolution_counter)));

      dbg D_trace (lazy ("o.instantiation_flag: "^(string_of_bool o.instantiation_flag)
                        ^" o.comb_inst_mult: "^(string_of_int o.comb_inst_mult)
                        ^" !instantiation_counter: "^(string_of_int !instantiation_counter)));
  *)

(*  
     if (o.resolution_flag &&
	 (!resolution_counter < o.comb_res_mult))
     then
*)
     if o.resolution_flag &&
	 ( (max_turn_deficit = !res_turn_deficit_ref) || !num_of_clauses_unchanged)
     then     
       (
        dbg D_trace (lazy ("resolution sec"));
	resolution_counter:= !resolution_counter + 1;
	(*num_resolution_loops := !num_resolution_loops+1;*)
	(try
	  assign_discount_time_limit (o.res_time_limit);
	  assign_discount_start_time ();
	  resolution_loop_exchange (ps_get_res_state ps_state);
          unassign_discount_time_limit ();
	with Timeout ->
	  unassign_discount_time_limit ();
	  out_str (pref_str^"Switching off resolution: loop timeout \n");
	  (if (not (o.instantiation_flag))
	  then 
	    (
	     out_warning "Switching on instantiation";
	     o.instantiation_flag <- true;
             inst_add_clause_list (ps_get_inst_state ps_state) input_clauses.ps_clauses_with_eq_axioms
	    )
	  ); 
	  if (o.instantiation_flag) then
	    (  o.resolution_flag <- false; 
(* TODO:       input_clauses_ref := simplify_input prover_functions !input_clauses_ref;*)
	       ps_res_clear ps_state;
	       clear_memory ()
	      )
	  (* switching_off_discount ()*)
	 (* else failwith "Inst is off and Resolution is TimeOut: see --res_time_limit" *)
	);
	(*       out_str ("Resolution flag: "^(string_of_bool !resolution_flag)^"\n");*)
	
	(* let f clause =
	   if (not (ClauseAssignDB.mem clause !clause_db_ref))
	   then ( clause)
	   else ()
	   in
	 *)
	(* List.iter
	   (Prop_solver_exchange.add_clause_to_solver gr_by) exchange_clauses;*)
	(* out_str ("\n Exchange Clauses: "
	   ^(Clause.clause_list_to_string exchange_clauses)^"\n") *)
       );
   (*  else *) (* end of resolution part *)
       (
(*
	if o.instantiation_flag &&
	  (!instantiation_counter < o.comb_inst_mult)
*)
        if o.instantiation_flag &&
	  ( (max_turn_deficit = !inst_turn_deficit_ref) ||  !num_of_clauses_unchanged)
	then
	  (
           dbg D_trace (lazy ("instantiation sec")); 
           instantiation_counter := !instantiation_counter + 1;
	   if ((not o.inst_learning_loop_flag) ||
	   (!learning_counter < !learning_bound))
	   then
	     (
	      learning_counter:=!learning_counter +1;
	      incr_int_stat 1 inst_num_of_loops;
              inst_lazy_loop_body (ps_get_inst_state ps_state);
	     )
	   else
	     (* learning: !current_options.inst_learning_loop_flag & !learning_counter >= !learning_bound *)
	     ((* Format.eprintf "Learning restart in instantiation@."; *)
	      learning_bound:=!learning_bound * o.inst_learning_factor;
	      learning_counter:=0;
	      
	      incr_int_stat 1 inst_num_of_learning_restarts;
(*	      (if (Prop_solver_exchange.solve ~reset:  true ()) = PropSolver.Unsat*)
              dbg D_solve (lazy "ps_full_loop");
              ps_reset_state ~opts ps_state;
	      (if Prop_solver_exchange.solve ~reset:false () == PropSolver.Unsat
	      then
		(* Raise separate exception, solver is not in an invalid state
		   and can be satisfiable without assumptions *)
		(* raise PropSolver.Unsatisfiable *)
		raise Unsatisfiable_gr
	      );
              
              (* run solve on solver_sim to infer more units in solver_sim *)

              if ((get_val_stat inst_num_of_learning_restarts) >= o.inst_start_prop_sim_after_learn)
              then
                (
                 let solver_sim_result =(Prop_solver_exchange.solve ~solver_in:Prop_solver_exchange.solver_sim ()) in 
                 assert (solver_sim_result != PropSolver.Unsat);
                 (* solver_sim should not be unsat if solver is not unsat *)
                );

              (if o.instantiation_flag then inst_add_clause_list (ps_get_inst_state ps_state) input_clauses.ps_clauses_with_eq_axioms);  
              (if o.resolution_flag    then res_add_clause_list (ps_get_res_state ps_state) input_clauses.ps_clauses_with_eq_axioms);  

 	      
(*              ps_clear_state ps_state; *)
	      (* simplify current_input_clauses with the new solver state *)
	      (* when simpl. given and new are switched off *)
	      (* not very good?  *)
	      (* let simplify_clause rest clause =
		 (Prop_solver_exchange.prop_subsumption clause):: rest
		 in
		 current_input_clauses := List.fold_left
		 simplify_clause [] !current_input_clauses;
	       *)
	      (* debug *)
	      (*
		out_str ("\n New Input Length: "
		^(string_of_int (List.length !current_input_clauses))^"\n");
	       *)
	      (* out_str "\nNon Horn Clauses:\n";
		 List.iter
		 (fun c ->
		 if (not (Clause.is_horn c))
		 then
		 out_str (Clause.to_string c)
		 else())
		 !current_input_clauses;
	       *)
	      (*		*)
	      (* out_str ("\n Simplifying input clauses after restart...");
		 let simp_input_clauses = List.map Prop_solver_exchange.prop_subsumption input_clauses in
		 out_str ("done\n");
	       *)
	      (* end debug *)
	      
(* TODO:      input_clauses_ref := simplify_input prover_functions !input_clauses_ref;  *)
(*
	      let module InstInput =
		struct
		  let inst_module_name =
		    ("Inst Restart "
		     ^(string_of_int (get_val_stat inst_num_of_learning_restarts)))
		      
		      (*		    let input_clauses = !current_input_clauses		  *)
		      (*		    let input_clauses = simp_input_clauses	       	*)
(*		  let input_clauses = !input_clauses_ref *)
		  let input_clauses = apply_fun prover_functions.inst_get_all_input_clauses ()
		end in
	      let module InstM = Instantiation.Make (InstInput) in
	      prover_functions.inst_lazy_loop_body <- Def(InstM.lazy_loop_body);
	      prover_functions.inst_add_clause <- Def(InstM.add_clause);
	      prover_functions.inst_get_all_input_clauses <- Def(InstM.get_all_input_clauses);
	      prover_functions.inst_clear_all <-Def(InstM.clear_all);
	      (*	      prover_functions.inst_learning_restart input_clauses;*)
*)
	      clear_memory ()
	     )
	  );
      (*  else *)
         ( if o.superposition_flag 
            &&
            ( (max_turn_deficit = !sup_turn_deficit_ref) || !num_of_clauses_unchanged)
         (*
           if o.superposition_flag 
            &&
	  (!superposition_counter < o.comb_sup_mult)
*)
        then (
           dbg D_trace (lazy ("superposition sec")); 
          incr superposition_counter;
          Superposition.step (ps_state.ps_sup_state |> Option.get);
         )
      (*  else (* instantiation_counter > instantiation_mult *)
          (
           resolution_counter:= 0;
           superposition_counter:=0;
           instantiation_counter:=0;
(* new *)
           failwith "ps_full_loop: a deadlock in turns";
          )*)            
       );
        dbg D_trace (lazy ("end sec")); 
    
(*        failwith "ps_full_loop: a deadlock in turns"; *)
      )   
         )
  done

(*-----------------new ps_full_loop (not used yet)----------------*)

let ps_full_loop_init ~opts ~time_limit input_clauses =
  let o = opts in  
  
  print_options_if_new ~opts;
  
  out_str "\n";
  print_string ((s_pref_str ())^"Proving...\n");
  flush stdout;

  let ps_state = ps_create_state ~input_clauses ~opts:o in
  dbg D_trace (lazy ("adding clauses to ps_state: "));
  dbg D_trace (lazy (Clause.clause_list_to_string input_clauses.ps_clauses_with_eq_axioms));

  (if o.instantiation_flag then inst_add_clause_list (ps_get_inst_state ps_state) input_clauses.ps_clauses_with_eq_axioms); 

  (if o.resolution_flag then res_add_clause_list (ps_get_res_state ps_state) input_clauses.ps_clauses_with_eq_axioms);  

  (if o.superposition_flag then 
    Superposition.add_input_clauses (ps_state.ps_sup_state |> Option.get) ((* EqualityTransformation.clauselist_to_eq *) input_clauses.ps_clauses_no_eq_axioms);
  );  
  dbg D_trace (lazy ("done adding clauses "));
  
  if (not o.resolution_flag) && (not o.instantiation_flag) && (not o.superposition_flag) then failwith "both --resolution_flag false and --instantiation_flag and --superposition_flag false";

  ps_state


(*--------- round robin functions ---------*)
(* functions excetuted at each iteration *)

let ps_loop_gen_fun ~opts state =
  let o = opts in
  if (not o.resolution_flag) && (not o.instantiation_flag) && (not o.superposition_flag) then failwith "both --resolution_flag false and --instantiation_flag and --superposition_flag false";

  prop_solver_per_cl_check ~opts:state.ps_opts
  

(*-------------*)
let ps_loop_inst ~opts ps_state =   
  let o = opts in

  if (not o.instantiation_flag)
  then ()
  else
    begin       
      if ((not o.inst_learning_loop_flag) ||
      (ps_state.ps_learning_counter < ps_state.ps_learning_bound))
      then
	(   
            ps_state.ps_inst_cnt <- ps_state.ps_inst_cnt +1;
            dbg D_trace (lazy ("ps_loop_inst: "^(string_of_int ps_state.ps_inst_cnt)));

	    ps_state.ps_learning_counter <- ps_state.ps_learning_counter +1;
	    incr_int_stat 1 inst_num_of_loops;
            inst_lazy_loop_body (ps_get_inst_state ps_state);
	   )
	   else
	     ((* Format.eprintf "Learning restart in instantiation@."; *)
	      ps_state.ps_learning_bound <- ps_state.ps_learning_bound * o.inst_learning_factor;
	      ps_state.ps_learning_counter <- 0;
	       
	      incr_int_stat 1 inst_num_of_learning_restarts;
 
(* reset only inst state *)
              ps_reset_inst_state ~opts ps_state;

              dbg D_solve (lazy "ps_loop_inst:learning restart"); 
	      (if Prop_solver_exchange.solve ~reset:false () == PropSolver.Unsat
	      then
		(* Raise separate exception, solver is not in an invalid state
		   and can be satisfiable without assumptions *)
		(* raise PropSolver.Unsatisfiable *)
		raise Unsatisfiable_gr
	      );
              
              (* run solve on solver_sim to infer more units in solver_sim *)

(* KK: clean! *)
              if ((get_val_stat inst_num_of_learning_restarts) >= o.inst_start_prop_sim_after_learn)
              then
                (
                 let solver_sim_result =(Prop_solver_exchange.solve ~solver_in:Prop_solver_exchange.solver_sim ()) in 
                 assert (solver_sim_result != PropSolver.Unsat);
                 (* solver_sim should not be unsat if solver is not unsat *)
                );
             )
    end


(*-------------*)
let ps_loop_res ~opts ps_state =   
  let o = opts in
  if (not o.resolution_flag)
  then ()
  else
    begin       
(*    (try *)
      (*KK: TODO move to local res_state! *)
(*
      assign_discount_time_limit (o.res_time_limit);
      assign_discount_start_time ();
*)
      ps_state.ps_res_cnt <- ps_state.ps_res_cnt +1;
      dbg D_trace (lazy ("ps_loop_res: "^(string_of_int ps_state.ps_res_cnt)));

      resolution_loop_exchange (ps_get_res_state ps_state);

(*      unassign_discount_time_limit (); *)
(*
    with Timeout ->
	  unassign_discount_time_limit ();
	  out_str (pref_str^"Switching off resolution: loop timeout \n");
	  (if (not (o.instantiation_flag))
	  then 
	    (
	     out_warning "Switching on instantiation";
	     o.instantiation_flag <- true;
             inst_add_clause_list (ps_get_inst_state ps_state) input_clauses.ps_clauses_with_eq_axioms
	    )
	  ); 
	  if (o.instantiation_flag) then
	    (  o.resolution_flag <- false; 
(* TODO:       input_clauses_ref := simplify_input prover_functions !input_clauses_ref;*)
	       ps_res_clear ps_state;
	       clear_memory ()
	      )
*)
      
    end

(*-------*)
let ps_loop_sup ~opts ps_state = 
  let o = opts in
  if not o.superposition_flag then 
    ()
  else (
    ps_state.ps_sup_cnt <- ps_state.ps_sup_cnt + 1;
    dbg D_trace @@ lazy (sprintf "ps_loop_sup: %d" ps_state.ps_sup_cnt);

    Superposition.step (ps_state.ps_sup_state |> Option.get);
  )



(* ================ new v2 =========*)

type inst_state_base = 
    {
     mutable inst_state : inst_state;
     mutable inst_learning_bound : int;
     mutable inst_learning_counter : int;
   }


type component_base = 
  |Inst of inst_state_base
  |Res of res_state 
  |Sup of Superposition.state
    
let comp_id_cnt = ref 0
    
type component = 
    { 
      comp                     : component_base; 
      comp_name                : string;
      comp_id                  : int;
      comp_opts                : Options.options;
      mutable comp_cnt         : int; (* how many times comp was called *)
      mutable last_deficit     : int; 
      mutable stagnation_cnt   : int; (* how many times comp was called without change in deficit *)
    }

type ps_state = 
    {
     ps_components     : component list; 
     ps_input_clauses  : clause list;
     ps_start_time     : float;
     ps_time_limit     : float; 
     ps_opts           : Options.options;
     ps_stagnation_limit : int;
     mutable ps_cnt    : int;
   }

type ps_comp_funs = 
    {
     ps_pre_comp_fun : ps_state -> unit;
     ps_post_comp_fun : ps_state -> unit;

     ps_comp_deficit_fun : ps_state -> comp -> int;
       (* ps_scheduler executes component with the largest deficit *)
       (* change to float ? *)

     ps_comp_main_fun : comp -> unit;    
       (* one iteration of component's reasoning loop *)          
       (* should also update state such that next call to deficit will be correctly computed *)

     ps_stagnation_fun : ps_state -> comp -> unit;
       (* stagnation occurs when deficit does not change for a component for ps_stagnation_limit calls of the comp. *)
       (* this can happen e.g. when component gets switched off (res limit) and does nothing after *)
   }
    
(*
type component = 
    {
     comp_deficit_fun : ps_state -> int;
         (* ps_scheduler executes component with the largest deficit *)
         (* change to float ? *)

     comp_main_fun : ps_state -> unit; 
         (* one iteration of component's reasoning loop *)          
         (* should also update state such that next call to deficit will be correctly computed *)

     mutable comp_loop_cnt : int; (* the number of calls to comp_main_fun *)
     comp_name : string; 
   }
*)

(*----------------------- ps_loop_main --------------*)

(* In a loop, from component list pick one with the maximal deficit and execute comp_main_fun *)
(* all *_fun are ps_state -> unit *)
(* comp_list -- component list *)
let ps_loop_main ps_state ~time_limit ps_state ps_comp_funs opts input_clauses = 

  let start_time = (Unix.gettimeofday ()) in  
  let o = opts in  
  
  let ps_state = ps_create_state ~input_clauses ~opts:o in
  init_fun ps_state; 
  
  while true 
  do
    check_sched_time ~start_time ~time_limit ps_state.ps_loop_cnt;
    pre_comp_fun ps_state;

    let comp_dfc_list = List.map (fun comp -> ((comp.comp_deficit_fun ps_state), comp)) comp_list in 
    let (max_dfc, max_dfc_comp) = list_find_max_element 
        (fun (dfc1,_comp1) (dfc2,_comp2) ->  Poly.compare dfc1 dfc2) 
        comp_dfc_list
    in 
    dbg D_trace (lazy (sprintf "ps_loop_cnt: %d; max_def_comp: %s; max deficit: %d" ps_state.ps_loop_cnt max_dfc_comp.comp_name max_dfc));
    
    max_dfc_comp.comp_main_fun ps_state; 
    
    post_comp_fun ps_state;   
    ps_state.ps_loop_cnt <- ps_state.ps_loop_cnt+1;
  done


(*-------------*)
let ps_init_fun ps_state = 

  print_options_if_new ~opts:ps_state.ps_opts;  
  out_str "\n";
  print_string ((s_pref_str ())^"Proving...\n");
  flush stdout


(*-------------*)
let ps_loop_inst ps_state =   
  let o = ps_state.opts in

  if (not o.instantiation_flag)
  then ()
  else
    begin       
      if ((not o.inst_learning_loop_flag) ||
      (ps_state.ps_learning_counter < ps_state.ps_learning_bound))
      then
	(   
            ps_state.ps_inst_cnt <- ps_state.ps_inst_cnt +1;
            dbg D_trace (lazy ("ps_loop_inst: "^(string_of_int ps_state.ps_inst_cnt)));

	    ps_state.ps_learning_counter <- ps_state.ps_learning_counter +1;
	    incr_int_stat 1 inst_num_of_loops;
            inst_lazy_loop_body (ps_get_inst_state ps_state);
	   )
	   else
	     ((* Format.eprintf "Learning restart in instantiation@."; *)
	      ps_state.ps_learning_bound <- ps_state.ps_learning_bound * o.inst_learning_factor;
	      ps_state.ps_learning_counter <- 0;
	       
	      incr_int_stat 1 inst_num_of_learning_restarts;
 
(* reset only inst state *)
              ps_reset_inst_state ~opts ps_state;

              dbg D_solve (lazy "ps_loop_inst:learning restart"); 
	      (if Prop_solver_exchange.solve ~reset:false () == PropSolver.Unsat
	      then
		(* Raise separate exception, solver is not in an invalid state
		   and can be satisfiable without assumptions *)
		(* raise PropSolver.Unsatisfiable *)
		raise Unsatisfiable_gr
	      );
              
              (* run solve on solver_sim to infer more units in solver_sim *)

(* KK: clean! *)
              if ((get_val_stat inst_num_of_learning_restarts) >= o.inst_start_prop_sim_after_learn)
              then
                (
                 let solver_sim_result =(Prop_solver_exchange.solve ~solver_in:Prop_solver_exchange.solver_sim ()) in 
                 assert (solver_sim_result != PropSolver.Unsat);
                 (* solver_sim should not be unsat if solver is not unsat *)
                );
             )
    end


(*------------ OLD------------------*)


(*
type prover_functions = {
    mutable inst_lazy_loop_body : (int ref -> int ref -> unit) param;
    mutable inst_add_clause : (clause -> unit) param;
    mutable inst_get_all_input_clauses : (unit -> clause list) param;
    mutable inst_clear_all : (unit -> unit) param;
    mutable res_discount_loop_exchange : (unit -> unit) param;
    mutable res_add_clause : (clause -> unit) param;
    mutable res_get_all_input_clauses : (unit -> clause list) param;
    mutable res_simplified_input : context param;
    mutable res_clear_all : (unit -> unit) param
  }
*)

(*
let provers_list = ref []


let provers_register pf =
  provers_list:= pf:: (!provers_list)
*)

(*
let clear_all_provers prover_functions =
  apply_fun_if_def prover_functions.inst_clear_all ();
  apply_fun_if_def prover_functions.res_clear_all ()
*)

(*
let provers_clear_and_remove_all () =
  List.iter 
    (fun pf -> 
      clear_all_provers pf; 
      pf.inst_lazy_loop_body <- Undef;
      pf.inst_add_clause <-Undef;
      pf.inst_get_all_input_clauses <-Undef;
      pf.inst_clear_all <- Undef;
      pf.res_discount_loop_exchange <- Undef; 
      pf.res_add_clause <- Undef;
      pf.res_get_all_input_clauses <-Undef;
      pf.res_simplified_input <- Undef;
      pf.res_clear_all <- Undef; 
    )
    !provers_list; 
  provers_list := []

let provers_clear_and_remove_top () =
  match !provers_list with
  | h:: tl ->
      clear_all_provers h;
      provers_list:= tl
  |[] -> ()

let get_inst_name_param_from_options () = 
  if !current_options.instantiation_flag 
  then
    Def("Inst")
  else 
    Undef		

let get_res_name_param_from_options () = 
  if !current_options.resolution_flag 
  then
    Def("Res")
  else 
    Undef		
      
      
let create_provers ~inst_name_param ~res_name_param input_clauses =
*)
(* reseting here is a problem; can have unsat without assumptions when it is unsat with assumptions...
 (if !current_options.reset_solvers
 then
   (Prop_solver_exchange.reset_solvers ())
 );
*)
(*
  let prover_functions = {
    inst_lazy_loop_body = Undef;
    inst_add_clause = Undef;
    inst_get_all_input_clauses = Undef;
    inst_clear_all = Undef;
    res_discount_loop_exchange = Undef;
    res_add_clause = Undef;
    res_get_all_input_clauses = Undef;
    res_simplified_input = Undef;
    res_clear_all = Undef
  } in
  (
   (*if !current_options.instantiation_flag 
     then*)
   match inst_name_param with 
   | Def(inst_name) ->
       let module InstInput =
	 struct
	   let inst_module_name = inst_name
	   let input_clauses = input_clauses
	 end in
       let module InstM = Instantiation.Make (InstInput) in
       prover_functions.inst_lazy_loop_body <- Def(InstM.lazy_loop_body);
       prover_functions.inst_add_clause <- Def(InstM.add_clause);
       prover_functions.inst_get_all_input_clauses <- Def(InstM.get_all_input_clauses);
       prover_functions.inst_clear_all <- Def(InstM.clear_all);
   | Undef ->	()
	 (*	else ()*)
  );
  ((*if !current_options.resolution_flag
     then *)
   match res_name_param with 
   | Def(res_name) ->
       
       let module ResInput =
	 struct
	   let res_module_name = res_name
	   let input_clauses = input_clauses
	   let is_res_prepocessing = false
	 end in
       let module ResM = Discount.Make (ResInput) in
       prover_functions.res_discount_loop_exchange <- Def(ResM.discount_loop_exchange);
       prover_functions.res_add_clause <- Def(ResM.add_clause);
       prover_functions.res_get_all_input_clauses <- Def(ResM.get_all_input_clauses);
       prover_functions.res_simplified_input <- Def(!ResM.simplified_input);
       prover_functions.res_clear_all <- Def(ResM.clear_all)
   | Undef ->	()
	 (*	else()*)
  );
  provers_register prover_functions;
  prover_functions

let create_provers_current_options input_clauses = 
  create_provers 
    ~inst_name_param:(get_inst_name_param_from_options ())
    ~res_name_param:(get_res_name_param_from_options ())
    input_clauses

*)

(* TODO: restore simplify_input *)  
(*
let simplify_input prover_functions clauses =
  if !current_options.res_sim_input
  then
  (  

(* TODO: experiment with preprocess!
     out_warning ("proof_search_loop: simplify_input -> preprocess \n\n");

   let prep_clauses =  
    Preprocess.preprocess ~is_eq_problem:(Problem_properties.has_equality clauses)  clauses in
   prep_clauses
*)
    match prover_functions.res_simplified_input with 
    | Def (simplified_input) -> 
	Clause.context_replace_by_clist simplified_input clauses
    |Undef -> clauses
)
  else 
    clauses 
*)

(*--- economical options output ---*)
